
//		mksc.cpp - Create, change, list, and read shortcut files


#define UNICODE						1
#define _UNICODE					1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR					1
#define VER_MINOR					4
#define VER_BUILD					4
#define VER_PATCH					5

#define PLUGIN_NAME					L"mksc"

#ifdef _WIN64
#define PLATFORM					L"x64"
#else
#define PLATFORM					L"x86"
#endif

#define MAX_ARG_LEN					8192
#define SHORT_BUF_LEN				1024
#define MAX_FILENAME_LEN			32768

#define ERRHANDLE					GetStdHandle( STD_ERROR_HANDLE )


#define MACRO_EXP( arg )			#arg
#define MACRO_STR( arg )			MACRO_EXP( arg )


#define RECURSE_NORMAL				1
#define RECURSE_NO_JUNCTIONS		2
#define RECURSE_NO_HIDDEN_DIRS		4
#define RECURSE_NO_SYSTEM_DIRS		8
#define RECURSE_NO_ERROR_MSGS		64

#define LF							0x0a
#define CR							0x0d
#define QUOTE						0x22


OSVERSIONINFOEXW OSVerInfo;


const wchar_t PluginFeatures[] = L"mksc,@scinfo";


HINSTANCE MyDLLHandle				= NULL;



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( !FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



#include "ParseArgs.cpp"
#include "NewHelp.cpp"



LPTSTR _PathPart( LPTSTR Filename, LPTSTR outPath )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 )
		return NULL;

	if ( outPath ) {
		for ( size_t i = 0; i <= EndOfPathPart; i++ )
			outPath[i] = Filename[i];

		outPath[EndOfPathPart + 1] = '\0';
		return outPath;
	}

	return Filename;
}



LPTSTR _FilenamePart( LPTSTR Filename, LPTSTR outFilename )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 ) {
		if ( outFilename ) {
			wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename );
			return outFilename;
		}
		else
			return Filename;
	}

	if ( outFilename ) {
		wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename + EndOfPathPart + 1 );
		return outFilename;
	}
	else
		return Filename + EndOfPathPart + 1;
}



LPTSTR _ExtensionPart( LPTSTR Filename, LPTSTR outExt )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	LPTSTR Ext			= NULL;


	if ( outExt )
		outExt[0] = '\0';

	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) || ( Filename[i] == ':' ) )
			Ext = NULL;
		else if ( Filename[i] == '.' )
			Ext = Filename + i;
	}

	if ( !Ext )
		return NULL;

	if ( outExt ) {
		wcscpy_s( outExt, MAX_FILENAME_LEN, Ext );
		return outExt;
	}

	return Ext;
}



BOOL HasWildcards( LPTSTR Filename )
{
	if ( !Filename )
		return FALSE;

	int start = 0;

	if ( _wcsnicmp( Filename, L"\\\\?\\", 4 ) == 0 )
		start = 4;

	for ( int i = start; Filename[i]; i++ )
		if ( ( Filename[i] == '?' ) || ( Filename[i] == '*' ) )
			return TRUE;

	return FALSE;
}



int StripQuotes( LPTSTR String )
{
	if ( !String )
		return -1;


	size_t i	= 0;
	size_t j	= 0;
	wchar_t ch	= '\0';
	int k		= 0;


	do {
		ch = String[i++];
		if ( ch == QUOTE ) {
			k++;
			continue;
		}

		String[j++] = ch;
	}
	while ( ch );
	
	return k;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion, LPCSTR SourceFile, LPCSTR SourceFileVer )
{
	wchar_t File[MAX_FILENAME_LEN]		= L"";
	wchar_t Ver[SHORT_BUF_LEN]			= L"";


	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion ) {
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

		if ( SourceFile && SourceFileVer ) {
			MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, SourceFile, -1, File, MAX_FILENAME_LEN );
			MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, SourceFileVer, -1, Ver, SHORT_BUF_LEN );
			StripQuotes( File );
			Printf( L"\t\t\t\t%s  %s\r\n", _FilenamePart( File, NULL), Ver );
		}
	}

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, PLATFORM );

	return 0;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	if ( ( ch == ':' ) || ( ch == '=' ) )
		return TRUE;
	else
		return FALSE;
}



int ReplaceChar( LPTSTR String, wchar_t FromChar, wchar_t ToChar )
{
	if ( !String )
		return -1;

	if ( !FromChar || !ToChar || ( FromChar == ToChar ) )
		return 0;


	int n = 0;


	for ( size_t i = 0; String[i]; i++ ) {
		if ( String[i] == FromChar ) {
			String[i] = ToChar;
			n++;
		}
	}

	return n;
}



BOOL IsAFile( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;
	else if ( Filename[0] == '\0' )
		return FALSE;


	DWORD attr		= GetFileAttributes( Filename );


	if ( attr == INVALID_FILE_ATTRIBUTES )
		return FALSE;

	if ( attr & ( FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_DEVICE ) )
		return FALSE;

	return TRUE;
}



int SearchAppPaths( LPTSTR inFilename, HKEY RootKey, LPTSTR outFullname )
{
	if ( !inFilename || !outFullname || !RootKey )
		return -1;
	else if ( inFilename[0] == '\0' )
		return -1;

	wchar_t KeyName[MAX_FILENAME_LEN]	= L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\";
	wchar_t Data[MAX_FILENAME_LEN]		= L"";
	HKEY KeyHandle = NULL;
	DWORD DataType = 0, DataSize = MAX_FILENAME_LEN * sizeof( wchar_t );
	int rv = 0;


	wcscat_s( KeyName, MAX_FILENAME_LEN, inFilename );

	if ( RegOpenKeyEx( RootKey, KeyName, 0, KEY_READ, &KeyHandle ) )
		return 1;

	rv = RegQueryValueEx( KeyHandle, NULL, NULL, &DataType, (BYTE *) Data, &DataSize );

	RegCloseKey( KeyHandle );

	if ( rv != ERROR_SUCCESS )
		return 1;

	if ( ( DataType != REG_SZ ) && ( DataType != REG_EXPAND_SZ ) )
		return 1;

	if ( ( DataType == REG_EXPAND_SZ ) && ( wcschr( Data, '%' ) != NULL ) ) {
		wchar_t temp[MAX_FILENAME_LEN] = L"";
		ExpandEnvironmentStrings( Data, temp, MAX_FILENAME_LEN );
		wcscpy_s( Data, MAX_FILENAME_LEN, temp );
	}

	if ( !IsAFile( Data ) )
		return 1;

	wcscpy_s( outFullname, MAX_FILENAME_LEN, Data );
	return 0;
}



int SearchPathForExe( LPTSTR inFilename, LPTSTR outFullname )
{
	if ( !inFilename || !outFullname )
		return -1;
	else if ( inFilename[0] == '\0' )
		return -1;

	outFullname[0] = '\0';

	if ( !SearchAppPaths( inFilename, HKEY_CURRENT_USER, outFullname ) )
		return 0;

	if ( !SearchAppPaths( inFilename, HKEY_LOCAL_MACHINE, outFullname ) )
		return 0;

	SearchPath( NULL, inFilename, NULL, MAX_FILENAME_LEN, outFullname, NULL );

	if ( outFullname[0] )
		return 0;

	return 1;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value	= 0;
	wchar_t *sep1		= NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return errno;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return errno;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



int ParseSignedInt( LPTSTR inString, int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return errno;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) || ( ( inString[0] == '-' ) && iswdigit( inString[1] ) ) ) {
		//  parse signed integer

		value = wcstol( inString, &sep1, 10 );

		if ( errno )
			return errno;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



int AppendFilename( LPTSTR Path, LPTSTR Filename, size_t BufferSize )
{
	if ( !Path || !Filename )
		return -1;

	
	BOOL NeedSlash		= FALSE;

	for ( size_t i = 0; Path[i]; i++ ) {
		if ( ( Path[i] == '/' ) || ( Path[i] == '\\' ) || ( Path[i] == ':' ) )
			NeedSlash = FALSE;
		else
			NeedSlash = TRUE;
	}

	if ( ( Filename[0] == '/' ) || ( Filename[0] == '\\' ) )
		NeedSlash = FALSE;
	
	if ( NeedSlash )
		wcscat_s( Path, BufferSize, L"\\" );

	wcscat_s( Path, BufferSize, Filename );
	return 0;
}



BOOL IsUnsafe( wchar_t ch )
{
	if ( ( ch == '%' ) || ( ch == '&' ) || ( ch == '(' ) || ( ch == ')' ) || ( ch == '<' ) || ( ch == '>' ) || ( ch == '[' ) || ( ch == ']' ) || ( ch == '^' ) || ( ch == '`' ) || ( ch == '|' ) )
		return TRUE;
	else
		return FALSE;
}



int MakeStringSafe( LPTSTR String )
{
	if ( !String )
		return -1;

	for ( int i = 0; String[i]; i++ ) {
		wchar_t ch = String[i];

		if ( IsUnsafe( ch ) )
			String[i] = ch + 0xfee0;
	}

	return 0;
}



#define MKSC_VARS_HELP
#include "MkScCmd.cpp"




BOOL GetVersion2( OSVERSIONINFOEXW *VerInfo )
{
	if ( !VerInfo )
		return -1;


	typedef NTSTATUS ( WINAPI _RtlGetVersionProc ) ( OSVERSIONINFOEXW *VerInfo );
	HMODULE NtDllHandle						= GetModuleHandle( L"NtDll.dll" );
	_RtlGetVersionProc *RtlGetVersionProc	= NULL;


	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	if ( NtDllHandle )
		RtlGetVersionProc = (_RtlGetVersionProc*) GetProcAddress( NtDllHandle, "RtlGetVersion" );

	if ( RtlGetVersionProc ) {
		RtlGetVersionProc( VerInfo );
		if ( VerInfo->dwMajorVersion >= 6 )
			return TRUE;
	}

	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	return ( GetVersionEx( (OSVERSIONINFOW*) VerInfo ) != 0 );
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
			MyDLLHandle = (HINSTANCE) hModule;
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/mksc.html";
	piInfo.pszDescription	= (LPTSTR) L"Create, modify, list, and read info from shortcut files.";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	GetVersion2( &OSVerInfo );


	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%u.%u.%u loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	CL_Initialize();
	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{

#ifdef NEWHELP_VERSION

	CloseHelpWindow();

#endif	//	 NEWHELP_VERSION

	CL_Shutdown();
	return 0;
}



const int NumHelpTexts							= 2;
const wchar_t HelpFeature[NumHelpTexts][11]		= { L"mksc", L"@scinfo" };
const wchar_t *HelpText[NumHelpTexts]			= { MkScHelpText, F_ScInfoHelpText };



#ifndef NEWHELP_VERSION

LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}

#endif	//	NEWHELP_VERSION



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}



#ifndef NEWHELP_VERSION

//		Simpler MessageBox() help, useful for smaller plugins which don't merit a full HTML help system:

DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}

#endif	//	NEWHELP_VERSION
